"""Defines the import for different source filetypes."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_input.ipynb.

# %% auto 0
__all__ = ['FocusAreaPathRestrictions', 'Data', 'Recording', 'ROI', 'DataLoader', 'GridWrapperROILoader', 'RecordingLoader',
           'AVILoader', 'ROILoader', 'ImageJROILoader', 'DataLoaderFactory', 'RecordingLoaderFactory',
           'ROILoaderFactory', 'get_filepaths_with_supported_extension_in_dirpath', 'RecLoaderROILoaderCombinator']

# %% ../nbs/03_input.ipynb 3
import imageio.v3 as iio
from pathlib import Path
import numpy as np
from shapely import Polygon
import roifile
from abc import ABC, abstractmethod
from typing import List, Dict, Tuple, Union, Optional, Any

# %% ../nbs/03_input.ipynb 4
class FocusAreaPathRestrictions:
    
    @property
    def supported_dir_names(self) -> List[str]:
        supported_dir_names =  ['focus_area', 'focus_areas', 'focus-area', 'focus-areas', 'focus area', 'focus areas',
                                'Focus_Area', 'Focus_Areas', 'Focus-Area', 'Focus-Areas', 'Focus Area', 'Focus Areas']
        return supported_dir_names

# %% ../nbs/03_input.ipynb 6
class Data(ABC):

    @abstractmethod
    def _parse_loaded_data(self, loaded_data: Any) -> None:
        pass

    def __init__(self, filepath: Path, loaded_data: Any) -> None:
        self.filepath = filepath
        self._parse_loaded_data(loaded_data)

# %% ../nbs/03_input.ipynb 7
class Recording(Data):

    def _parse_loaded_data(self, loaded_data: np.ndarray) -> None:
        self.zstack = loaded_data
        self.estimated_bit_depth = self._estimate_bit_depth()
        self.preview = self._create_brightness_and_contrast_enhanced_preview()


    def _estimate_bit_depth(self) -> int:
        max_bit_value = self.zstack.max()
        if max_bit_value <= 255:
            estimated_bit_depth = 255
        elif max_bit_value <= 4095:
            estimated_bit_depth = 4095
        elif max_bit_value <= 65535:
            estimated_bit_depth = 65535
        else:
            raise ValueError(f'Max bit value in recording found to be {max_bit_value}, but NA3 currently only handles up to 16-bit recordings!')
        return estimated_bit_depth


    def _create_brightness_and_contrast_enhanced_preview(self, percentile_for_adjustment: int=1) -> np.ndarray:
        raw_image = self.zstack[0, :, :, :].copy() # ensure that dimensions are the same as for ".get_single_frame_as_preview()"
        lower_percentile_bit_value = np.percentile(raw_image, percentile_for_adjustment)
        upper_percentile_bit_value = np.percentile(raw_image, 100-percentile_for_adjustment)
        contrast_adjustment_factor = self.estimated_bit_depth / (upper_percentile_bit_value - lower_percentile_bit_value)
        brightness_adjustment_factor = -(contrast_adjustment_factor * lower_percentile_bit_value)
        raw_image_clipped_at_percentile_bit_values = self._clip_image_at_bit_values(raw_image, lower_percentile_bit_value, upper_percentile_bit_value)
        brightness_contrast_adjusted_image = contrast_adjustment_factor * raw_image_clipped_at_percentile_bit_values + brightness_adjustment_factor
        return brightness_contrast_adjusted_image
        

    def _compute_contrast_and_brightness_adjustment_factors(self, raw_image: np.ndarray, percentile_for_adjustment: int=1) -> Tuple[float, float]:
        lower_percentile_bit_value = np.percentile(raw_image, percentile_for_adjustment)
        upper_percentile_bit_value = np.percentile(raw_image, 100-percentile_for_adjustment)
        contrast_adjustment = self.estimated_bit_depth / (upper_percentile_bit_value - lower_percentile_bit_value)
        brightness_adjustment = -(contrast_adjustment * lower_percentile_bit_value)
        return contrast_adjustment, brightness_adjustment


    def _clip_image_at_bit_values(self, raw_image: np.ndarray, min_bit_value: float, max_bit_value: float) -> np.ndarray:
        raw_image[raw_image <= min_bit_value] = min_bit_value
        raw_image[raw_image >= max_bit_value] = max_bit_value
        return raw_image

# %% ../nbs/03_input.ipynb 8
class ROI(Data):

    def _parse_loaded_data(self, loaded_data: List[Tuple[int, int]]) -> None:
        self.boundary_row_col_coords = loaded_data
        self.as_polygon = self._convert_to_valid_polygon()


    def _convert_to_valid_polygon(self) -> Polygon:
        roi_as_polygon = Polygon(self.boundary_row_col_coords)
        assert roi_as_polygon.is_valid, f'Something went wrong when trying to create a Polygon out of your ROI: {self.filepath}.'
        return roi_as_polygon


    def add_label_id(self, label_id: str) -> None:
        assert type(label_id) == str, f'"label_id" must be a string. However, you passed {label_id} which is of type {type(label_id)}.'
        setattr(self, 'label_id', label_id)

# %% ../nbs/03_input.ipynb 10
class DataLoader(ABC):

    @abstractmethod
    def load_and_parse_file_content(self) -> Union[Data, List[Data]]:
        # This method will be called when the data should be loaded for analysis
        pass

    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath

# %% ../nbs/03_input.ipynb 11
class GridWrapperROILoader(DataLoader):

    def set_configs_for_grid_creation(self, image_width: int, image_height: int, window_size: int) -> None:
        self.configs = {}
        self._add_to_configs_and_create_as_attribute('image_width', image_width)
        self._add_to_configs_and_create_as_attribute('image_height', image_height)
        self._add_to_configs_and_create_as_attribute('window_size', window_size)


    def _add_to_configs_and_create_as_attribute(self, attribute_name: str, value: Any) -> None:
        self.configs[attribute_name] = value
        setattr(self, attribute_name, self.configs[attribute_name])

    
    def load_and_parse_file_content(self) -> List[ROI]:
        row_cropping_idx, col_cropping_idx = self._get_cropping_indices_to_adjust_for_window_size()
        self._add_to_configs_and_create_as_attribute('row_cropping_idx', row_cropping_idx)
        self._add_to_configs_and_create_as_attribute('col_cropping_idx', col_cropping_idx)        
        grid_row_idxs, grid_col_idxs = self._get_row_col_idxs_of_grid()
        grid_row_labels, grid_col_labels = self._get_row_col_labels_for_rois_in_grid()
        self._add_to_configs_and_create_as_attribute('max_len_row_label_id', len(str(grid_row_labels[-1])))
        self._add_to_configs_and_create_as_attribute('max_len_col_label_id', len(str(grid_col_labels[-1])))
        all_rois = []
        for row_idx, row_label in zip(grid_row_idxs, grid_row_labels):
            for col_idx, col_label in zip(grid_col_idxs, grid_col_labels):
                square_corner_row_col_coords = self._get_boundary_row_col_coords_single_square(row_idx, col_idx)
                label_id = f'{row_label}/{col_label}'
                square_roi = ROI(self.filepath, square_corner_row_col_coords)
                square_roi.add_label_id(label_id)
                all_rois.append(square_roi)
        return all_rois     

    
    def _get_cropping_indices_to_adjust_for_window_size(self) -> Tuple[int, int]:
        row_cropping_index = (self.image_height // self.window_size) * self.window_size
        col_cropping_index = (self.image_width // self.window_size) * self.window_size
        return row_cropping_index, col_cropping_index
        

    def _get_row_col_idxs_of_grid(self) -> Tuple[np.ndarray, np.ndarray]:
        grid_row_idxs = np.arange(start = 0, stop = self.row_cropping_idx, step = self.window_size)
        grid_col_idxs = np.arange(start = 0, stop = self.col_cropping_idx, step = self.window_size)
        return grid_row_idxs, grid_col_idxs


    def _get_row_col_labels_for_rois_in_grid(self) -> Tuple[np.ndarray, np.ndarray]:
        grid_row_labels = np.arange(start = 1, stop = self.row_cropping_idx / self.window_size + 1, step = 1, dtype = 'int')
        grid_col_labels = np.arange(start = 1, stop = self.col_cropping_idx / self.window_size + 1, step = 1, dtype = 'int')
        return grid_row_labels, grid_col_labels
                
    
    def _get_boundary_row_col_coords_single_square(self, upper_left_corner_row_idx: int, upper_left_corner_col_idx: int) -> List[Tuple[(int, int)]]:
        upper_left_corner = (upper_left_corner_row_idx, upper_left_corner_col_idx)
        upper_right_corner = (upper_left_corner_row_idx, upper_left_corner_col_idx + self.window_size)
        lower_right_corner = (upper_left_corner_row_idx + self.window_size, upper_left_corner_col_idx + self.window_size)
        lower_left_corner = (upper_left_corner_row_idx + self.window_size, upper_left_corner_col_idx)
        return [upper_left_corner, lower_left_corner, lower_right_corner, upper_right_corner, upper_left_corner]

# %% ../nbs/03_input.ipynb 12
class RecordingLoader(DataLoader):


    @abstractmethod
    def _get_all_frames(self) -> np.ndarray: 
        # To be implemented in individual subclasses
        # Shape of returned numpy array: [frames, rows, cols, color_channels]
        pass


    def _load_all_frames(self) -> np.ndarray: 
        all_frames = self._get_all_frames()
        all_frames = self._validate_shape_and_convert_to_grayscale_if_possible(all_frames)
        return all_frames


    def load_and_parse_file_content(self) -> Recording:
        all_frames = self._load_all_frames()
        recording = Recording(self.filepath, all_frames)
        return recording


    def _validate_shape_and_convert_to_grayscale_if_possible(self, zstack: np.ndarray) -> np.ndarray:
        self._validate_correct_array_shape(zstack)
        if zstack.shape[3] > 1:
            if self._check_if_color_channels_are_redunant(zstack) == True:
                zstack = self._convert_to_grayscale(zstack)
        return zstack


    def _validate_correct_array_shape(self, zstack: np.ndarray) -> None:
        assert len(zstack.shape) == 4, ('The shape of the zstack numpy array is not correct. It should be a 4 dimensional array, like '
                                        f'[frames, rows, cols, color channels]. However, the current shape is: {zstack.shape}.')
        assert zstack.shape[3] in [1, 3], ('The color channels of the recording you attempted to load are incorrect. Currently, only single '
                                           f'channel or RGB (i.e. 1 or 3 color channels) are supported. However, your data has: {zstack.shape[3]}.')
        
    
    def _check_if_color_channels_are_redunant(self, zstack: np.ndarray) -> bool:
        reference_channel_idx = 0
        color_channels_are_equal = []
        for idx_of_channel_to_compare in range(1, zstack.shape[3]):
            if np.array_equal(zstack[:, :, :, reference_channel_idx], zstack[: , :, :, idx_of_channel_to_compare]) == True:
                color_channels_are_equal.append(True)
            else:
                color_channels_are_equal.append(False)
                break
        return all(color_channels_are_equal)


    def _convert_to_grayscale(self, zstack: np.ndarray) -> np.ndarray:
        return zstack[:, :, :, 0:1]

# %% ../nbs/03_input.ipynb 13
class AVILoader(RecordingLoader):

    def _get_all_frames(self) -> np.ndarray: 
        return iio.imread(self.filepath)

# %% ../nbs/03_input.ipynb 14
class ROILoader(DataLoader):

    @abstractmethod
    def _get_boundary_row_col_coords_for_all_rois_in_source_data(self) -> List[List[Tuple[int, int]]]: 
        # To be implemented in individual subclasses
        # Return a list of Tuples, where each tuple represents one boundary point: (row_coord, col_coord)
        pass

    
    def load_and_parse_file_content(self) -> List[ROI]:
        boundary_row_col_coords_for_all_rois = self._get_boundary_row_col_coords_for_all_rois_in_source_data()
        all_rois = []
        for boundary_row_col_coords_single_roi in boundary_row_col_coords_for_all_rois:
            boundary_row_col_coords_single_roi = self._add_first_boundary_point_also_add_end_to_close_roi(boundary_row_col_coords_single_roi)
            roi = ROI(self.filepath, boundary_row_col_coords_single_roi)
            all_rois.append(roi)
        return all_rois


    def _add_first_boundary_point_also_add_end_to_close_roi(self, boundary_row_col_coords: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
        first_boundary_point_coords = boundary_row_col_coords[0]
        boundary_row_col_coords.append(first_boundary_point_coords)
        return boundary_row_col_coords

# %% ../nbs/03_input.ipynb 15
class ImageJROILoader(ROILoader):

    
    def _get_boundary_row_col_coords_for_all_rois_in_source_data(self) -> List[List[Tuple[int, int]]]:
        roi_file_content = roifile.roiread(self.filepath)
        if type(roi_file_content) == list:
            all_rois = self._extract_boundary_row_col_coords_from_roi_set(roi_file_content)
        else:
            all_rois = [self._extract_boundary_row_col_coords_from_single_roi(roi_file_content)]
        return all_rois


    def _extract_boundary_row_col_coords_from_roi_set(self, all_imagej_rois: List[roifile.roifile.ImagejRoi]) -> List[List[Tuple[int, int]]]:
        boundary_coords_all_rois = []
        for imagej_roi in all_imagej_rois:
            boundary_row_col_coords_single_roi = self._extract_boundary_row_col_coords_from_single_roi(imagej_roi)
            boundary_coords_all_rois.append(boundary_row_col_coords_single_roi)
        return boundary_coords_all_rois
    
        
    def _extract_boundary_row_col_coords_from_single_roi(self, imagej_roi: roifile.roifile.ImagejRoi) -> List[Tuple[int, int]]:
        row_coords = imagej_roi.coordinates()[:, 1]
        col_coords = imagej_roi.coordinates()[:, 0]
        boundary_row_col_coords = list(zip(row_coords, col_coords))
        return boundary_row_col_coords

# %% ../nbs/03_input.ipynb 17
class DataLoaderFactory(ABC):

    @property
    @abstractmethod
    def supported_extensions_per_data_loader(self) -> Dict[DataLoader, List[str]]:
        pass

    
    @property
    def all_supported_extensions(self) -> List[str]:
        all_supported_extensions = []
        for value in self.supported_extensions_per_data_loader.values():
            all_supported_extensions += value
        return all_supported_extensions

    
    def get_loader(self, filepath: Path) -> DataLoader:
        self._assert_validity_of_filepath(filepath)
        data_loader = self._get_loader_for_file_extension(filepath)
        return data_loader
        

    def _assert_validity_of_filepath(self, filepath: Path) -> None:
        assert isinstance(filepath, Path), f'filepath must be an instance of a pathlib.Path. However, you passed {filepath}, which is of type {type(filepath)}'
        assert filepath.exists(), f'The filepath you provided ({filepath}) does not seem to exist!'
        

    def _get_loader_for_file_extension(self, filepath: Path) -> DataLoader:
        matching_loader = None
        for loader_subclass, supported_extensions in self.supported_extensions_per_data_loader.items():
            if filepath.suffix in supported_extensions:
                matching_loader = loader_subclass(filepath)
                break
        if matching_loader == None:
            raise NotImplementedError('It seems like there is no DataLoader implemented for the specific filetype you´re trying to load - sorry!')
        return matching_loader

# %% ../nbs/03_input.ipynb 18
class RecordingLoaderFactory(DataLoaderFactory):

    @property
    def supported_extensions_per_data_loader(self) -> Dict[RecordingLoader, List[str]]:
        supported_extensions_per_data_loader = {AVILoader: ['.avi']}
        return supported_extensions_per_data_loader

# %% ../nbs/03_input.ipynb 19
class ROILoaderFactory(DataLoaderFactory):

    @property
    def supported_extensions_per_data_loader(self) -> Dict[ROILoader, List[str]]:
        supported_extensions_per_data_loader = {ImageJROILoader: ['.roi', '.zip']}
        return supported_extensions_per_data_loader

# %% ../nbs/03_input.ipynb 20
def get_filepaths_with_supported_extension_in_dirpath(dirpath: Path, all_supported_extensions: List[str], max_results: Optional[int]=None) -> List[Path]:
    all_filepaths_with_supported_extension = []
    for elem in dirpath.iterdir():
        if elem.is_file() == True:
            if elem.suffix in all_supported_extensions:
                all_filepaths_with_supported_extension.append(elem)
    if type(max_results) == int:
        assert len(all_filepaths_with_supported_extension) <= max_results, (
            f'There are more than {max_results} file(s) of supported type in {dirpath}, '
            f'but only a maximum of {max_results} are allowed. Please remove at least '
            f'{len(all_filepaths_with_supported_extension) - max_results} of the following' 
            f'files and try again: {all_filepaths_with_supported_extension}'
        )
    return all_filepaths_with_supported_extension

# %% ../nbs/03_input.ipynb 21
class RecLoaderROILoaderCombinator:

        
    def __init__(self, dir_path: Path) -> None:
        self.dir_path = dir_path

    
    def get_all_recording_and_roi_loader_combos(self) -> List[Tuple[RecordingLoader, ROILoader]]:
        recording_loader = self._get_the_recording_loader()
        all_roi_loaders = self._get_all_roi_loaders()
        if len(all_roi_loaders) > 0:
            rec_roi_loader_combos = [(recording_loader, roi_loader) for roi_loader in all_roi_loaders]
        else:
            rec_roi_loader_combos = [(recording_loader, None)]
        return rec_roi_loader_combos
    

    def _get_the_recording_loader(self) -> RecordingLoader:
        recording_loader_factory = RecordingLoaderFactory()
        recording_filepath = get_filepaths_with_supported_extension_in_dirpath(self.dir_path, recording_loader_factory.all_supported_extensions, 1)[0]
        recording_loader = recording_loader_factory.get_loader(recording_filepath)
        return recording_loader


    def _get_all_roi_loaders(self) -> List[ROILoader]:
        roi_loader_factory = ROILoaderFactory()
        all_roi_filepaths = get_filepaths_with_supported_extension_in_dirpath(self.dir_path, roi_loader_factory.all_supported_extensions)
        all_roi_loaders = [roi_loader_factory.get_loader(filepath) for filepath in all_roi_filepaths]
        return all_roi_loaders



"""Defines the structure of individual analysis jobs."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_processing.ipynb.

# %% auto 0
__all__ = ['process_analysis_rois', 'AnalysisJob']

# %% ../nbs/04_processing.ipynb 3
import multiprocessing
import numpy as np
import pandas as pd
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes 
from datetime import datetime, timezone
from shapely import get_coordinates
from abc import ABC, abstractmethod

# %% ../nbs/04_processing.ipynb 4
from typing import Optional, Tuple, Dict, List, Any, Union

# %% ../nbs/04_processing.ipynb 5
from .input import DataLoader, RecordingLoader, Recording, ROILoader, ROI, GridWrapperROILoader
from . import results
from .analysis import AnalysisROI

# %% ../nbs/04_processing.ipynb 6
def process_analysis_rois(analysis_roi: AnalysisROI, configs: Dict[str, Any]) -> AnalysisROI:
    analysis_roi.compute_mean_intensity_timeseries(configs['limit_analysis_to_frame_interval'], configs['start_frame_idx'], configs['end_frame_idx'])
    if np.mean(analysis_roi.mean_intensity_over_time) >= configs['signal_average_threshold']:
        analysis_roi.detect_peaks(configs['signal_to_noise_ratio'], configs['octaves_ridge_needs_to_spann'], configs['noise_window_size'])
        analysis_roi.estimate_baseline(configs['baseline_estimation_method'])
        analysis_roi.compute_area_under_curve()
        analysis_roi.compute_amplitude_and_delta_f_over_f()
        analysis_roi.compute_variance_area(configs['variance'])
    return analysis_roi

# %% ../nbs/04_processing.ipynb 7
class AnalysisJob:

    def __init__(self, 
                 number_of_parallel_processes: int,
                 data_loaders: Dict[str, Union[DataLoader, List[DataLoader]]]
                ) -> None:
        self.number_of_parallel_processes = number_of_parallel_processes
        self.recording_loader = data_loaders['recording']
        self.parent_dir_path = self.recording_loader.filepath.parent  
        if 'rois' in data_loaders.keys():
            self.rois_source = 'file'
            self.roi_loaders = data_loaders['rois']
        else:
            self.rois_source = 'grid'
            self.roi_loaders = [GridWrapperROILoader(self.recording_loader.filepath)]
        if 'focus_area' in data_loaders.keys():
            self.focus_area_enabled = True
            self.focus_area_loader = data_loaders['focus_area']
        else:
            self.focus_area_enabled = False
            self.focus_area = None

    
    def preview_window_size(self, window_size: int) -> Tuple[Figure, Axes]:
        self.load_data_into_memory(window_size)
        grid_configs = self.roi_loaders[0].configs
        fig, ax = results.plot_window_size_preview(self.recording.preview, grid_configs, self.focus_area)
        return fig, ax

    
    def load_data_into_memory(self, window_size: int) -> None:
        if hasattr(self, 'recording') == False:
            self.recording = self.recording_loader.load_and_parse_file_content()
        if hasattr(self, 'all_analysis_rois') == False:
            self.all_analysis_rois = self._create_all_analysis_rois(window_size)
        elif self.rois_source == 'grid':
            if self.roi_loaders[0].window_size != window_size:
                self.all_analysis_rois = self._create_all_analysis_rois(window_size)


    def _create_all_analysis_rois(self, window_size: int) -> List[AnalysisROI]:     
        self.all_rois = self._load_data_from_all_roi_loaders(window_size)
        if self.focus_area_enabled == True:
            self.focus_area = self.focus_area_loader.load_and_parse_file_content()[0]
            self.all_rois = self._filter_rois_by_focus_area()
        return self._create_analysis_rois()

    
    def _load_data_from_all_roi_loaders(self, window_size: int) -> List[ROI]:
        all_rois = []
        for roi_loader in self.roi_loaders:
            if type(roi_loader) == GridWrapperROILoader:
                roi_loader.set_configs_for_grid_creation(self.recording.preview.shape[1], self.recording.preview.shape[0], window_size)
            all_rois += roi_loader.load_and_parse_file_content()
        if self.rois_source == 'grid':
            self.grid_configs = self.roi_loaders[0].configs
        else:
            self.grid_configs = None
        return all_rois
    

    def _filter_rois_by_focus_area(self) -> List[ROI]:
        filtered_rois = [roi for roi in self.all_rois if roi.as_polygon.within(self.focus_area.as_polygon)]
        return filtered_rois
        

    def _create_analysis_rois(self) -> List[AnalysisROI]:
        all_analysis_rois = []
        if self.rois_source == 'file':
            self._create_and_add_label_ids_to_all_rois_from_file()
        for roi in self.all_rois:
            roi_bounding_box_row_col_coords = get_coordinates(roi.as_polygon.envelope).astype('int')
            row_min, col_min = roi_bounding_box_row_col_coords.min(axis=0)
            row_max, col_max = roi_bounding_box_row_col_coords.max(axis=0)   
            zstack = self.recording.zstack[:, row_min:row_max, col_min:col_max, :]
            analysis_roi = AnalysisROI(roi, (row_min, col_min), zstack)
            all_analysis_rois.append(analysis_roi)
        return all_analysis_rois


    def _create_and_add_label_ids_to_all_rois_from_file(self) -> List[str]:
        roi_count = len(self.all_rois)
        zfill_factor = int(np.log10(roi_count)) + 1
        for idx, roi in enumerate(self.all_rois):
            label_id = str(idx + 1).zfill(zfill_factor)
            roi.add_label_id(label_id)

    
    def run_analysis(self,
                     window_size: int,
                     limit_analysis_to_frame_interval: bool,
                     start_frame_idx: int,
                     end_frame_idx: int,
                     signal_average_threshold: float,
                     signal_to_noise_ratio: float,
                     octaves_ridge_needs_to_spann: float,
                     noise_window_size: int,
                     baseline_estimation_method: str,                     
                     include_variance: bool,
                     variance: int
                    ) -> None:
        self._set_analysis_start_datetime()
        self.load_data_into_memory(window_size)
        configs = locals()
        configs.pop('self')
        copy_of_all_analysis_rois = self.all_analysis_rois.copy()
        with multiprocessing.Pool(processes = self.number_of_parallel_processes) as pool:
            processed_analysis_rois = pool.starmap(process_analysis_rois, [(analysis_roi, configs) for analysis_roi in copy_of_all_analysis_rois])
        self.all_analysis_rois = processed_analysis_rois


    def _set_analysis_start_datetime(self) -> None:
            users_local_timezone = datetime.now().astimezone().tzinfo
            self.analysis_start_datetime = datetime.now(users_local_timezone) 


    def create_results(self, 
                       save_overview_png: bool,
                       save_detailed_results: bool,
                       minimum_activity_counts: int, 
                       signal_average_threshold: float, 
                       signal_to_noise_ratio: float
                      ) -> None:
        self._ensure_results_dir_exists()
        activity_filtered_analysis_rois = [roi for roi in self.all_analysis_rois if roi.peaks_count >= minimum_activity_counts]
        
        self.activity_overview_plot = results.plot_activity_overview(analysis_rois_with_sufficient_activity = activity_filtered_analysis_rois,
                                                                     preview_image = self.recording.preview,
                                                                     indicate_activity = True,
                                                                     focus_area = self.focus_area,
                                                                     grid_configs = self.grid_configs)
        if save_overview_png == True:
            self.activity_overview_plot[0].savefig(self.results_dir_path.joinpath('activity_overview.png'), dpi = 300)
            label_id_overview_fig, label_id_overview_ax = results.plot_rois_with_label_id_overview(analysis_rois_with_sufficient_activity = activity_filtered_analysis_rois,
                                                                                                   preview_image = self.recording.preview,
                                                                                                   focus_area = self.focus_area,
                                                                                                   grid_configs = self.grid_configs)
            label_id_overview_fig.savefig(self.results_dir_path.joinpath('ROI_label_IDs_overview.png'), dpi = 300)
            plt.close()
        if save_detailed_results == True:
            self._create_and_save_csv_result_files(activity_filtered_analysis_rois)
            self._create_and_save_individual_traces_pdf_result_file(activity_filtered_analysis_rois)


    def _ensure_results_dir_exists(self) -> None:
        if hasattr(self, 'results_dir_path') == False:
            prefix_with_datetime = self.analysis_start_datetime.strftime('%Y_%m_%d_%H-%M-%S_results_for')
            recording_filename_without_extension = self.recording.filepath.name.replace(self.recording.filepath.suffix, '')
            if self.focus_area_enabled == True:
                focus_area_filename_without_extension = self.focus_area.filepath.name.replace(self.focus_area.filepath.suffix, '')
                results_dir_name = f'{prefix_with_datetime}_{recording_filename_without_extension}_with_{focus_area_filename_without_extension}'                
            else:
                results_dir_name = f'{prefix_with_datetime}_{recording_filename_without_extension}'
            self.results_dir_path = self.parent_dir_path.joinpath(results_dir_name)
            self.results_dir_path.mkdir()       


    def _create_variance_area_dataframe(self, filtered_rois: List[AnalysisROI]) -> pd.DataFrame:
        data = {'ROI label ID': [],
                'Variance Area': []}
        for roi in filtered_rois:
            data['ROI label ID'].append(roi.label_id)
            data['Variance Area'].append(roi.variance_area)
        df = pd.DataFrame(data = data)
        return df
    
    
    def _create_and_save_csv_result_files(self, filtered_rois: List[AnalysisROI]) -> None:
        if len(filtered_rois) > 0:
            df_variance_areas = self._create_variance_area_dataframe(filtered_rois)
            df_variance_areas.to_csv(self.results_dir_path.joinpath('Variance_area_results.csv'), index = False)
            peak_results_per_roi = [results.export_peak_results_df_from_analysis_roi(roi) for roi in filtered_rois]
            df_all_peak_results = pd.concat(peak_results_per_roi, ignore_index = True)
            max_peak_count_across_all_rois = df_all_peak_results.groupby('ROI label ID').count()['peak frame index'].max()
            zfill_factor = int(np.log10(max_peak_count_across_all_rois)) + 1
            amplitude_and_delta_f_over_f_results_all_rois = []
            auc_results_all_rois = []
            for roi_label_id in df_all_peak_results['ROI label ID'].unique():
                tmp_df_single_roi = df_all_peak_results[df_all_peak_results['ROI label ID'] == roi_label_id].copy()
                amplitude_and_delta_f_over_f_results_all_rois.append(results.create_single_roi_amplitude_and_delta_f_over_f_results(tmp_df_single_roi, zfill_factor))
                auc_results_all_rois.append(results.create_single_roi_auc_results(tmp_df_single_roi, zfill_factor))
            df_all_amplitude_and_delta_f_over_f_results = pd.concat(amplitude_and_delta_f_over_f_results_all_rois, ignore_index = True)
            df_all_auc_results = pd.concat(auc_results_all_rois, ignore_index = True)
            # Once all DataFrames are created successfully, write them to disk 
            df_all_peak_results.to_csv(self.results_dir_path.joinpath('all_peak_results.csv'), index = False)
            df_all_amplitude_and_delta_f_over_f_results.to_csv(self.results_dir_path.joinpath('Amplitude_and_dF_over_F_results.csv'), index = False)
            df_all_auc_results.to_csv(self.results_dir_path.joinpath('AUC_results.csv'), index = False)

    
    def _create_and_save_individual_traces_pdf_result_file(self, filtered_rois: List[AnalysisROI]) -> None:
            filepath = self.results_dir_path.joinpath('Individual_traces_with_identified_events.pdf')
            with PdfPages(filepath) as pdf:
                for indicate_activity in [True, False]:
                    overview_fig, ax = results.plot_activity_overview(filtered_rois, self.recording.preview, indicate_activity, self.focus_area, self.grid_configs)
                    pdf.savefig(overview_fig)
                    plt.close()
                label_ids_overview_fig, ax = results.plot_rois_with_label_id_overview(filtered_rois, self.recording.preview, self.focus_area, self.grid_configs)
                pdf.savefig(label_ids_overview_fig)
                plt.close()
                for roi in filtered_rois:
                    fig = results.plot_intensity_trace_with_identified_peaks_for_individual_roi(roi)
                    pdf.savefig(fig)
                    plt.close()



import ipywidgets as w
from ipyfilechooser import FileChooser
from IPython.display import Image
import numpy as np
from datetime import datetime

from typing import Dict, Optional, Any
from pathlib import Path


def change_widget_state(widget: w.Widget,
                        value: Optional[Any]=None,
                        description: Optional[str]=None,
                        disabled: Optional[bool]=None,
                        visibility: Optional[str]=None,
                        tooltip: Optional[str]=None,
                        button_style: Optional[str]=None
                       ) -> w.Widget:
    if value != None:
        widget.value = value
    if description != None:
        widget.description = description
    if disabled != None:
        widget.disabled = disabled
    if visibility != None:
        widget.layout.visibility = visibility
    if tooltip != None:
        widget.tooltip = tooltip
    if button_style != None:
        widget.button_style = button_style
    return widget




class UserInfoPanel:

    def __init__(self) -> None:
        self.widget = self._build_widget()
        # Initialize basic parameters and configs:
        self.user_timezone = datetime.now().astimezone().tzinfo
        self.max_lines_that_can_be_displayed_in_output = int(self.detailed_logs_output.layout.max_height.replace('px', '')) / 15
        self.logs_message_count = 0
        self.progress_in_percent = 0.0 # still required?


    def _build_widget(self) -> w.VBox:
        info = w.HTML(value="<p style='font-size:16px; font-weight:bold; text-align:center;'>Info</p>", layout = w.Layout(width = '40px'))
        self.latest_logs = w.Label(value = ' ... the latest logs message will be displayed here ... ',
                                   style = {'font_style': 'italic', 'text_color': 'gray', 'font_family': 'monospace', 'text_align': 'center'},
                                   layout = w.Layout(width = 'initial'))
        self.progress_bar = w.FloatProgress(description = 'Progress', style = {'description_width': 'initial'}, layout = w.Layout(width = '20%'))
        info_overview_box = w.HBox([info, self.latest_logs, self.progress_bar], layout = w.Layout(justify_content = 'space-between', align_items = 'center', width = '95%'))
        self.detailed_logs_output = w.Output(layout = w.Layout(max_height = '200px', y_overflow='scroll'))
        self.detailed_logs_accordion = w.Accordion(children = [self.detailed_logs_output], titles = ('Detailed logs', ), selected_index = None, layout = w.Layout(width = '95%'))
        vertical_spacer = w.HTML(value = '', layout = w.Layout(height = '15px'))
        return w.VBox([info_overview_box, self.detailed_logs_accordion, vertical_spacer], layout = w.Layout(align_items = 'center'))


    def add_new_logs(self, message: str) -> None:
        assert type(message) == str, f'UserInfoHandler.add_new_logs() expects a message of type string. However, you passed {message}, which is of type {type(message)}.'
        current_time = datetime.now(self.user_timezone)
        # Update latest logs:
        time_prefix_latest_logs = current_time.strftime('%H-%M-%S')
        if len(message) > 125:
            self.latest_logs.value = f'{time_prefix_latest_logs}: {message[:85]}...(find full message in detailed logs)'
        else:
            self.latest_logs.value = f'{time_prefix_latest_logs}: {message}'
        # Update detailed logs:
        time_prefix_detailed_logs = current_time.strftime('%d-%m-%y %H:%M:%S.%f')
        self.logs_message_count += 1
        # Check if title has already been changed:
        if self.detailed_logs_accordion.get_title(0) == 'Detailed logs':
            # If not, check if max lines have been reached with this new message and adjust title accordingly:
            if self.logs_message_count >= self.max_lines_that_can_be_displayed_in_output:
                self.detailed_logs_accordion.set_title(0, 'Detailed logs - please scroll down to see latest logs')
        with self.detailed_logs_output:
            print(f'{time_prefix_detailed_logs}: {message}')


    def update_progress_bar(self, progress_in_percent: float) -> None:
        assert type(progress_in_percent) == float, ('UserInfoPanel.update_progress_bar() expects the progress_in_percent of type float. '
                                                    f'However, you passed {progress_in_percent}, which is of type {type(progress_in_percent)}.')
        if progress_in_percent == 100.0:
            self.progress_bar.bar_style = 'success'
        else:
            self.progress_bar.bar_style = ''
        self.progress_bar.value = progress_in_percent



class SourceDataPanel:
    
    def __init__(self, user_info_panel: UserInfoPanel) -> None:
        self.user_info_panel = user_info_panel
        self.widget = self._build_widget()

    
    def _build_widget(self) -> w.HBox:
        # Create and configure all elements:
        io_source_data_info = w.HTML(value="<p style='font-size:16px; font-weight:bold; text-align:center;'>General Settings</p>")
        self.user_settings_batch_mode = w.Checkbox(description = 'Enable batch processing', value = False, indent = False)
        self.user_settings_focus_area_enabled = w.Checkbox(description = 'Limit analysis to focus area', value = False, indent = False)
        self.user_settings_roi_mode = w.Dropdown(options = [('Use adjustable grid to create ROIs (congruent squares)', 'grid'), ('Load predefined ROIs from source data', 'file')],
                                                 description = 'ROIs:',
                                                 value = 'grid',
                                                 style = {'description_width': 'initial'}, 
                                                 layout = {'width': 'initial'})
        processing_modes_box = w.VBox([self.user_settings_roi_mode, w.HBox([self.user_settings_batch_mode, self.user_settings_focus_area_enabled])], layout = {'width': '33%'})
        self.user_settings_data_source_path = FileChooser(title = 'Please select the recording file:', layout = w.Layout(width = '50%'))
        self.user_settings_data_source_path.rows = 8
        self.load_source_data_button = w.Button(description = 'Load Data', 
                                                disabled = True, 
                                                tooltip = 'Please first select which source data to load!', 
                                                layout = w.Layout(width = '12%', height = '55px'))
        vertical_spacer = w.HTML(value = '', layout = w.Layout(height = '5px'))
        # Enable event handling:
        self.user_settings_batch_mode.observe(self._change_batch_mode_config)
        self.user_settings_focus_area_enabled.observe(self._change_focus_area_config)
        self.user_settings_roi_mode.observe(self._change_roi_mode_config)
        self.user_settings_data_source_path.register_callback(self._data_source_path_chosen)
        # Arrange elements:
        source_data_settings_box = w.HBox([processing_modes_box,
                                           self.user_settings_data_source_path,
                                           self.load_source_data_button], 
                                          layout = w.Layout(width = '100%', justify_content = 'space-between'))
        general_settings_box = w.VBox([io_source_data_info, source_data_settings_box, vertical_spacer], layout = w.Layout(width = '95%', align_items = 'center'))
        return w.HBox([general_settings_box], layout = w.Layout(width = '100%', justify_content = 'center'))


    def _change_user_settings_data_source_path_configs(self, show_dirs_only: bool, title: str, reset: bool) -> None:
        if reset == True:
            self.user_settings_data_source_path.reset()
            self.load_source_data_button = change_widget_state(self.load_source_data_button, disabled = True, tooltip = 'Please select which source data to load!')
        self.user_settings_data_source_path.show_only_dirs = show_dirs_only
        self.user_settings_data_source_path.title = title
        
    
    def _change_roi_mode_config(self, change) -> None:
        if change['name'] == 'value':
            if change['new'] == 'file':
                show_only_dirs = True
                if self.user_settings_batch_mode.value == True:
                    title = 'Please select the parent directory that contains subdirectories with the individual source data:'
                elif self.user_settings_focus_area_enabled.value == True:
                    title = 'Please select the directory that contains the recording, all ROI files, and a directory with focus area ROI(s):'
                else:
                    title = 'Please select the directory that contains the recording and all ROI files:'
            else:
                if self.user_settings_batch_mode.value == True:
                    title = 'Please select the parent directory that contains subdirectories with the individual source data:'
                    show_only_dirs = True
                elif self.user_settings_focus_area_enabled.value == True:
                    title = 'Please select the directory that contains the recording and a directory with focus area ROI(s):'
                    show_only_dirs = True
                else:
                    title = 'Please select the recording file:'
                    show_only_dirs = False
            reset = self.user_settings_data_source_path.show_only_dirs != show_only_dirs
            self._change_user_settings_data_source_path_configs(show_only_dirs, title, reset)
           
        

    def _change_focus_area_config(self, change) -> None:
        if change['name'] == 'value':
            roi_mode = self.user_settings_roi_mode.value
            fake_change = {'name': 'value', 'new': roi_mode}
            self._change_roi_mode_config(fake_change)
           

    
    def _change_batch_mode_config(self, change) -> None:
        if change['name'] == 'value':
            roi_mode = self.user_settings_roi_mode.value
            fake_change = {'name': 'value', 'new': roi_mode}
            self._change_roi_mode_config(fake_change)                     


    def _data_source_path_chosen(self, file_chooser_obj) -> None:
        if file_chooser_obj.value != None:
            enable_loading = False
            source_data_path = Path(file_chooser_obj.value)
            if self.user_settings_roi_mode.value == 'file':
                if source_data_path.is_dir() == True:
                    enable_loading = True
            else:
                if (self.user_settings_batch_mode.value == True) or (self.user_settings_focus_area_enabled.value == True):
                    if source_data_path.is_dir() == True:
                        enable_loading = True
                else:
                    if source_data_path.is_file() == True:
                        enable_loading = True
            if enable_loading == True:
                self.load_source_data_button = change_widget_state(self.load_source_data_button, disabled = False, tooltip = 'Click to load the selected source data')
            else:
                self.load_source_data_button = change_widget_state(self.load_source_data_button, disabled = True, tooltip = 'Please select which source data to load!')


class AnalysisSettingsPanel:

    def __init__(self) -> None:
        self.widget = self._build_default_widget()


    def _build_default_widget(self) -> None:
        width_percentage_core_widgets = '95%'
        description_width = '35px'
        # Create and configure all elements:
        analysis_settings_info = w.HTML(value="<p style='font-size:16px; font-weight:bold; text-align:center;'>Analysis Settings</p>", layout = w.Layout(width = '99%'))

        self.user_settings_window_size = w.IntSlider(description = 'Grid size:', value = 10, min = 1, max = 128, step = 1, disabled = True, 
                                                          layout = w.Layout(width = '80%'), style = {'description_width': 'initial'})
        self.preview_window_size_button = w.Button(description = 'Preview', disabled = True, tooltip = 'Preview grid size. Does not start analysis', layout = w.Layout(width = '20%'))
        self.user_settings_signal_to_noise_ratio = w.BoundedFloatText(description = 'SNR: ', tooltip = 'Signal to noise ratio', value = 3.0, min = 0.1, max = 100.0, step = 0.05, disabled = True, 
                                                                      layout = w.Layout(width = width_percentage_core_widgets), style = {'description_width': description_width})
        self.user_settings_noise_window_size = w.BoundedIntText(description = 'NWS: ', tooltip = 'Noise window size', value = 200, min = 10, max = 1000, step = 1, disabled = True,
                                                                layout = w.Layout(width = width_percentage_core_widgets), style = {'description_width': description_width})
        self.user_settings_signal_average_threshold = w.BoundedFloatText(description = 'SAT: ', tooltip = 'Signal average threshold', value = 10.0, min = 0.0, max = 255.0, step = 0.5, 
                                                                         disabled = True, layout = w.Layout(width = width_percentage_core_widgets), 
                                                                         style = {'description_width': description_width})
        self.user_settings_minimum_activity_counts = w.BoundedIntText(description = 'MAC: ', tooltip = 'Minimum activity counts', value = 2, min = 0, max = 100, step = 1, disabled = True, 
                                                                      layout = w.Layout(width = width_percentage_core_widgets), style = {'description_width': description_width})
        self.user_settings_baseline_estimation_method = w.Dropdown(description = 'Baseline estimation method: ',
                                                                   value = 'asls', 
                                                                   options = [("Asymmetric Least Squares", "asls"),
                                                                              ("Fully Automatic Baseline Correction", "fabc"),
                                                                              ("Peaked Signal's Asymmetric Least Squares Algorithm", "psalsa"),
                                                                              ("Standard Deviation Distribution", "std_distribution")
                                                                             ],
                                                                   disabled = True,
                                                                   layout = w.Layout(width = '99%'), style = {'description_width': 'initial'})
        vertical_spacer = w.HTML(value = '', layout = w.Layout(height = '4px'))
        horizontal_spacer = w.HTML(value = '', layout = w.Layout(width = '5%'))
        vbox_core_settings_left = w.VBox([self.user_settings_signal_to_noise_ratio, vertical_spacer, self.user_settings_signal_average_threshold], 
                                         layout = w.Layout(width = '50%', align_items = 'flex-start'))
        vbox_core_settings_right = w.VBox([self.user_settings_noise_window_size, vertical_spacer, self.user_settings_minimum_activity_counts], 
                                           layout = w.Layout(width = '50%', align_items = 'flex-end'))      
        hbox_grid_window_size_settings = w.HBox([self.user_settings_window_size, self.preview_window_size_button], layout = w.Layout(width = '100%', justify_content = 'flex-start'))
        hbox_core_setting_vboxes = w.HBox([vbox_core_settings_left, vbox_core_settings_right], layout = w.Layout(width = '100%'))
        vbox_all_core_settings = w.VBox([hbox_grid_window_size_settings, 
                                         vertical_spacer,
                                         hbox_core_setting_vboxes,
                                         vertical_spacer,
                                         self.user_settings_baseline_estimation_method], 
                                        layout = w.Layout(width = '90%'))

        
        dashed_separator_line = w.HTML(value = "<hr style='border: none; border-bottom: 1px dashed;'>", layout = w.Layout(width = '95%'))
        optional_info = w.Label(value = 'Optional Settings:', style = {'text_align': 'left', 'font_weight': 'bold'}, layout = w.Layout(width = '90%'))
        self.user_settings_include_variance = w.Checkbox(description = 'include variance', value = False, indent = False, disabled = True, 
                                                         layout = w.Layout(width = '35%'), style = {'description_width': 'initial'})
        self.user_settings_variance = w.BoundedIntText(description = 'Variance:', disabled = True,
                                                         value = 15, min = 5, max = 200, step = 5,
                                                         style = {'description_width': 'initial'},
                                                         layout = w.Layout(width = '65%', visibility = 'hidden'))
        self.user_settings_limit_analysis_to_frame_interval = w.Checkbox(description = 'analyze interval', indent = False, 
                                                                         value = False, disabled = True, layout = w.Layout(width = '35%'), 
                                                                         style = {'description_width': 'initial'})
        self.user_settings_frame_interval_to_analyze = w.IntRangeSlider(description = 'Frames:', disabled = True, 
                                                                        value = (0, 500), min = 0, max = 500, step = 1, 
                                                                        style = {'description_width': 'initial'}, layout = w.Layout(width = '65%', visibility = 'hidden'))
        self.user_settings_configure_octaves = w.Checkbox(description = 'configure octaves', value = False, disabled = True, indent = False, 
                                                          layout = w.Layout(width = '35%'), style = {'description_width': 'initial'})
        self.user_settings_octaves_ridge_needs_to_spann = w.BoundedFloatText(description = 'Min. octaves:', tooltip = 'Minimum octaves a ridge needs to span', disabled = True,
                                                                             value = 1.0, min = 0.1, max = 30.0, step = 0.05,
                                                                             style = {'description_width': 'initial'},
                                                                             layout = w.Layout(width = '65%', visibility = 'hidden'))
        optional_variance_widgets = w.HBox([self.user_settings_include_variance, self.user_settings_variance], layout = w.Layout(width = '100%', align_items = 'flex-start'))
        optional_interval_widgets = w.HBox([self.user_settings_limit_analysis_to_frame_interval, self.user_settings_frame_interval_to_analyze], 
                                           layout = w.Layout(width = '100%', align_items = 'flex-start'))
        optional_octave_widgets = w.HBox([self.user_settings_configure_octaves, self.user_settings_octaves_ridge_needs_to_spann], 
                                         layout = w.Layout(width = '100%', align_items = 'flex-start'))
        
        optional_settings = w.VBox([optional_variance_widgets, vertical_spacer, optional_interval_widgets, vertical_spacer, optional_octave_widgets],
                                     layout = w.Layout(width = '90%', align_items = 'flex-start', align_content = 'flex-start', justify_content = 'flex-start'))

        results_info = w.Label(value = 'Results Settings:', style = {'text_align': 'left', 'font_weight': 'bold'}, layout = w.Layout(width = '90%'))
        self.user_settings_save_overview_png = w.Checkbox(description = 'Save overview plot', value = True, disabled = True, style = {'description_width': 'initial'})
        self.user_settings_save_detailed_results = w.Checkbox(description = 'Save detailed results', value = True, disabled = True, style = {'description_width': 'initial'})
        self.run_analysis_button = w.Button(description = 'Run Analysis',
                                            disabled = True,
                                            tooltip = 'You have to load some data first, before you can run the analysis!',
                                            button_style = '',
                                            icon = 'rocket',
                                            layout = w.Layout(width = '90%'))
        results_settings = w.VBox([w.HBox([self.user_settings_save_overview_png, self.user_settings_save_detailed_results],
                                          layout = w.Layout(width = '100%', align_items = 'flex-start'))],
                                   layout = w.Layout(width = '90%', align_items = 'flex-start', align_content = 'flex-start', justify_content = 'flex-start'))
        # Enable event handling:
        self.user_settings_include_variance.observe(self._include_variance_config_changed)
        self.user_settings_limit_analysis_to_frame_interval.observe(self._limit_analysis_to_interval_changed)
        self.user_settings_configure_octaves.observe(self._configure_octaves_changed)
        # Arrange elements:
        analysis_settings_box = w.VBox([analysis_settings_info,
                                        vbox_all_core_settings,
                                        vertical_spacer,
                                        dashed_separator_line, 
                                        optional_info,
                                        optional_settings,
                                        vertical_spacer,
                                        dashed_separator_line,
                                        results_info,
                                        results_settings,
                                        vertical_spacer,
                                        self.run_analysis_button,
                                        vertical_spacer],
                                       layout = w.Layout(height = '512px', width = '33%', align_items = 'center', border_top = '1px solid', border_bottom = '1px solid'))
        return analysis_settings_box

    
    def enable_analysis_settings(self, enable_all_widgets: bool, roi_mode: str) -> None:
        if enable_all_widgets == True:
            self.run_analysis_button = change_widget_state(self.run_analysis_button, 
                                                           disabled = False, 
                                                           button_style = 'success',
                                                           tooltip = 'Click here to start the analysis with all currently specified settings!')
        else:
            self.run_analysis_button = change_widget_state(self.run_analysis_button, 
                                                           disabled = True,
                                                           button_style = '')     
        for attribute_name, attribute_obj in vars(self).items():
            if attribute_name.startswith('user_settings'):
                attribute_obj.disabled = not enable_all_widgets
            elif attribute_name == 'preview_window_size_button':
                attribute_obj.disabled = not enable_all_widgets
        if roi_mode == 'file':
            self.preview_window_size_button.disabled = True
            self.user_settings_window_size.disabled = True
    

    def _include_variance_config_changed(self, change) -> None:
        if change['name'] == 'value':
            if change['new'] == True:
                self.user_settings_variance.layout.visibility = 'visible'
            else:
                self.user_settings_variance.layout.visibility = 'hidden'


    def _limit_analysis_to_interval_changed(self, change) -> None:
        if change['name'] == 'value':
            if change['new'] == True:
                self.user_settings_frame_interval_to_analyze.layout.visibility = 'visible'
            else:
                self.user_settings_frame_interval_to_analyze.layout.visibility = 'hidden'


    def _configure_octaves_changed(self, change) -> None:
        if change['name'] == 'value':
            if change['new'] == True:
                self.user_settings_octaves_ridge_needs_to_spann.layout.visibility = 'visible'
            else:
                self.user_settings_octaves_ridge_needs_to_spann.layout.visibility = 'hidden'





class MainScreen:

    def __init__(self) -> None:
        self.welcome_screen = self._build_welcome_screen()
        self.output_screen = self._build_output_screen()
        self.widget = w.VBox([], layout = w.Layout(height = '512px', width = '67%', justify_content = 'center', align_items = 'center', 
                                                                border_top = '1px solid', border_left = '1px solid', border_bottom = '1px solid'))
        self.show_welcome_screen()

    def _build_welcome_screen(self) -> w.VBox:
        welcome = w.HTML(value="<p style='font-size:32px; font-weight:bold; text-align:center;'>Welcome to</p>", layout = w.Layout(width = '99%'))
        logo = w.HTML(value='<img src="https://raw.githubusercontent.com/jpits30/NeuronActivityTool/master/Logo.png" width="256" height="256">')
        start_instructions = w.HTML(value="<p style='font-size:20px; font-weight:bold; text-align:center;'>Please start by selecting data of a recording to analyze</p>", layout = w.Layout(width = '99%'))
        welcome_screen = w.VBox([welcome, logo, start_instructions])
        return welcome_screen


    def _build_output_screen(self) -> w.VBox:
        self.output = w.Output()
        output_screen = w.VBox([self.output])
        return output_screen


    def show_welcome_screen(self) -> None:
        self.widget.children = self.welcome_screen.children
        self.current_screen = 'welcome'


    def show_output_screen(self, clear_output: bool=True) -> None:
        if clear_output == True:
            self.output.clear_output()
        self.widget.children = self.output_screen.children
        self.current_screen = 'output'



class WidgetsInterface:

    def __init__(self) -> None:
        self.user_info_panel = UserInfoPanel()
        self.source_data_panel = SourceDataPanel(user_info_panel = self.user_info_panel)
        self.analysis_settings_panel = AnalysisSettingsPanel()
        self.main_screen = MainScreen()
        self.widget = w.VBox([self.source_data_panel.widget, 
                              w.HBox([self.analysis_settings_panel.widget, self.main_screen.widget]),
                              self.user_info_panel.widget], layout = w.Layout(align_items = 'stretch', border = '1px solid'))
        self._setup_observer_for_roi_mode_config_change()


    def _setup_observer_for_roi_mode_config_change(self) -> None:
        self.source_data_panel.user_settings_roi_mode.observe(self._enable_window_size_widgets)


    def _enable_window_size_widgets(self, change) -> None:
        if change['name'] == 'value':
            if change['new'] == 'file':
                change_widget_state(self.analysis_settings_panel.user_settings_window_size, disabled = True)
                change_widget_state(self.analysis_settings_panel.preview_window_size_button, disabled = True)
            else:
                if self.analysis_settings_panel.user_settings_signal_to_noise_ratio.disabled == True:
                    pass
                else:
                    change_widget_state(self.analysis_settings_panel.user_settings_window_size, disabled = False)
                    change_widget_state(self.analysis_settings_panel.preview_window_size_button, disabled = False)                


    def update_infos(self, logs_message: Optional[str]=None, progress_in_percent: Optional[float]=None) -> None:
        if logs_message != None:
            self.user_info_panel.add_new_logs(logs_message)
        if progress_in_percent != None:
            self.user_info_panel.update_progress_bar(progress_in_percent)


    def adjust_widgets_to_loaded_data(self, total_frames: int) -> None:
        self._adjust_frame_interval_selection_widget(total_frames)


    def _adjust_frame_interval_selection_widget(self, total_frames: int) -> None:
        self.analysis_settings_panel.user_settings_frame_interval_to_analyze.max = total_frames + 1
        self.analysis_settings_panel.user_settings_frame_interval_to_analyze.value = (1, total_frames + 1)


    def export_user_settings(self) -> Dict[str, Any]:
        user_settings = {}
        for panel_name_with_user_settings in ['analysis_settings_panel', 'source_data_panel']:
            panel = getattr(self, panel_name_with_user_settings)
            for attribute_name, attribute_value in vars(panel).items():
                if attribute_name == 'user_settings_frame_interval_to_analyze':
                    user_settings['start_frame_idx'] = attribute_value.value[0]
                    user_settings['end_frame_idx'] = attribute_value.value[1]
                elif attribute_name.startswith('user_settings_'):
                    value_set_by_user = attribute_value.value
                    if value_set_by_user != None:
                        parameter_name = attribute_name.replace('user_settings_', '')
                        if 'path' in parameter_name:
                            value_set_by_user = Path(value_set_by_user)
                        user_settings[parameter_name] = value_set_by_user               
        return user_settings


    def enable_analysis(self, enable: bool=True) -> None:
        roi_mode = self.source_data_panel.user_settings_roi_mode.value
        self.analysis_settings_panel.enable_analysis_settings(enable, roi_mode)     


import matplotlib.pyplot as plt
from matplotlib.transforms import Bbox
import numpy as np
import pandas as pd
from shapely import get_coordinates
from typing import List, Tuple, Optional, Union, Any, Dict, Never
from pathlib import Path
from matplotlib.text import Text
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes 

from .input import ROI
from .analysis import AnalysisROI



def plot_roi_boundaries(roi: Union[ROI, AnalysisROI], line_color: str, line_style: str, line_width: Union[int, float]) -> None:
    roi_boundary_row_col_coords = np.asarray(roi.boundary_row_col_coords)
    plt.plot(roi_boundary_row_col_coords[:, 1], roi_boundary_row_col_coords[:, 0], c = line_color, linestyle = line_style, linewidth = line_width)


def _use_grid_row_col_labels_as_axis_labels(image_shape: Tuple[int, int], grid_configs: Dict[str, Any], ax: Axes, font_size: int=12) -> None:
    ax.set_xticks(np.arange(0, image_shape[1], grid_configs['window_size']), labels = [])
    ax.set_xticks(np.arange(grid_configs['window_size']/2, grid_configs['col_cropping_idx'] + grid_configs['window_size']/2, grid_configs['window_size']), 
                  labels = np.arange(1, grid_configs['col_cropping_idx']/grid_configs['window_size'] + 1, 1, dtype='int'),
                  minor = True,
                  fontsize = min(12, font_size))
    ax.xaxis.set_label_text('X')
    ax.set_yticks(np.arange(0,  image_shape[0], grid_configs['window_size']), labels = [])
    ax.set_yticks(np.arange(grid_configs['window_size']/2, grid_configs['row_cropping_idx'] + grid_configs['window_size']/2, grid_configs['window_size']), 
                  labels = np.arange(1, grid_configs['row_cropping_idx']/grid_configs['window_size'] + 1, 1, dtype='int'), 
                  minor = True, 
                  fontsize = min(12, font_size))
    ax.yaxis.set_label_text('Y')
    ax.tick_params(bottom = False, left = False)


def _plot_grid(grid_configs: Dict[str, Any], 
               grid_line_color: str,
               grid_line_style: str,
               grid_line_width: int,
               focus_area_roi: Union[ROI, None],
               area_line_color: str,
               area_line_style: str,
               area_line_width: int,
               ax: Axes
              ) -> None:
    ax.grid(color = grid_line_color, linestyle = grid_line_style, linewidth = grid_line_width)
    if focus_area_roi == None:
        plt.hlines([0, grid_configs['row_cropping_idx']], 
                   xmin = 0, 
                   xmax = grid_configs['col_cropping_idx'], 
                   color = area_line_color,
                   linestyle = area_line_style,
                   linewidth = area_line_width)
        plt.vlines([0, grid_configs['col_cropping_idx']],
                   ymin = 0, 
                   ymax = grid_configs['row_cropping_idx'], 
                   color = area_line_color,
                   linestyle = area_line_style,
                   linewidth = area_line_width)
    else:
        plot_roi_boundaries(focus_area_roi, area_line_color, area_line_style, area_line_width)    



def plot_window_size_preview(preview_image: np.ndarray, 
                             grid_configs: Dict[str, Any], 
                             focus_area_roi: Optional[ROI]=None
                            ) ->Tuple[Figure, Axes]:
    fig, ax = plt.subplots()
    ax.imshow(preview_image, cmap="gray") # vmin = np.percentile(preview_image, 2.5), vmax = np.percentile(preview_image, 97.5)
    _plot_grid(grid_configs, 'gray', 'dashed', 1, focus_area_roi, 'cyan', 'solid', 2, ax)
    _use_grid_row_col_labels_as_axis_labels(preview_image.shape, grid_configs, ax) 
    ax.set_title(f'Preview of grid for window size: {grid_configs["window_size"]}')
    return fig, ax


def _get_text_bounding_box_size_in_data_dimensions(text: Text, fig: Figure, ax: Axes) -> Tuple[float, float]:
    renderer = fig.canvas.get_renderer()
    text_bbox_raw_dimensions = text.get_window_extent(renderer=renderer)
    text_bbox_data_dimensions = Bbox(ax.transData.inverted().transform(text_bbox_raw_dimensions))
    return np.abs(text_bbox_data_dimensions.width), np.abs(text_bbox_data_dimensions.height)


def _iteratively_decrease_fontsize_to_fit_text_in_squares(text: Text, max_size: float, fig: Figure, ax: Axes) -> None:
    text.set_fontsize(text.get_fontsize()-1)
    text_width, text_height = _get_text_bounding_box_size_in_data_dimensions(text, fig, ax)
    if (text_width > max_size) or (text_height > max_size): 
        _iteratively_decrease_fontsize_to_fit_text_in_squares(text, max_size, fig, ax)


def _get_fontsize_adjusted_to_slimmest_roi(preview_image: np.ndarray, min_width: int, max_text_length: int, default_fontsize: int=25) -> float:
    max_text_size = min_width*0.75
    tmp_fig, tmp_ax = plt.subplots()
    tmp_ax.imshow(preview_image)
    sample_coord = min_width + 0.5 * min_width
    max_width_text_from_number = '4'*max_text_length
    tmp_text = tmp_ax.text(sample_coord, sample_coord, max_width_text_from_number, fontsize = default_fontsize)
    text_width, text_height = _get_text_bounding_box_size_in_data_dimensions(tmp_text, tmp_fig, tmp_ax)
    if (text_width > max_text_size) or (text_height > max_text_size):
        _iteratively_decrease_fontsize_to_fit_text_in_squares(tmp_text, max_text_size, tmp_fig, tmp_ax)
    adjusted_fontsize = tmp_text.get_fontsize()
    plt.close(tmp_fig)
    return adjusted_fontsize


def _get_min_width_among_all_rois(analysis_rois: List[AnalysisROI], grid_configs: Optional[Dict[str, Any]]) -> int:
    if grid_configs != None:
        min_width = grid_configs['window_size']
    else:
        all_roi_widths = []
        for roi in analysis_rois:
            bounding_box_col_coords = get_coordinates(roi.as_polygon.envelope)[:, 1]
            min_col_idx = bounding_box_col_coords.min()
            max_col_idx = bounding_box_col_coords.max()
            width = max_col_idx - min_col_idx
            all_roi_widths.append(width)
        min_width = min(all_roi_widths)
    return min_width


def _get_all_peak_counts(analysis_rois: List[AnalysisROI]) -> List[int]:
    all_peak_counts = [roi.peaks_count for roi in analysis_rois]
    return all_peak_counts


def _get_max_peak_count(analysis_rois: List[AnalysisROI]) -> int:
    all_peak_counts = _get_all_peak_counts(analysis_rois)
    if len(all_peak_counts) > 0:
        max_peak_count = max(all_peak_counts)
    else:
        max_peak_count = 0
    return max_peak_count  


def _get_total_peak_count(analysis_rois: List[AnalysisROI]) -> int:
    all_peak_counts = _get_all_peak_counts(analysis_rois)
    if len(all_peak_counts) == 0:
        total_peak_count = 0
    else:
        total_peak_count = sum(all_peak_counts)
    return total_peak_count


def _plot_text_at_roi_centroid_coordinates(analysis_roi: AnalysisROI, text: str, size: int, color: str, ax: Axes) -> None:
        ax.text(analysis_roi.centroid_row_col_coords[1], 
                analysis_roi.centroid_row_col_coords[0], 
                text, 
                color = color, 
                horizontalalignment='center', 
                verticalalignment = 'center', 
                fontsize = size)    


def _plot_peak_count_text_for_each_analysis_roi(analysis_rois: List[AnalysisROI], preview_image: np.ndarray, grid_configs: Optional[Dict[str, Any]], ax: Axes) -> None:
    max_peak_count = _get_max_peak_count(analysis_rois)
    max_len_peak_count_text = len(str(max_peak_count))
    min_width_among_all_rois = _get_min_width_among_all_rois(analysis_rois, grid_configs)
    font_size = _get_fontsize_adjusted_to_slimmest_roi(preview_image, min_width_among_all_rois, max_len_peak_count_text)
    for roi in analysis_rois:
        _plot_text_at_roi_centroid_coordinates(roi, roi.peaks_count, font_size, 'magenta', ax)


def plot_activity_overview(analysis_rois_with_sufficient_activity: Union[List[AnalysisROI], List[Never]],
                           preview_image: np.ndarray,
                           indicate_activity: bool=False,
                           focus_area: Optional[ROI]=None,
                           grid_configs: Optional[Dict[str, Any]]=None
                          ) -> Tuple[Figure, Axes]:
    fig, ax = plt.subplots()
    ax.imshow(preview_image, cmap="gray")
    if len(analysis_rois_with_sufficient_activity) > 0:
        if indicate_activity == True:
            _plot_peak_count_text_for_each_analysis_roi(analysis_rois_with_sufficient_activity, preview_image, grid_configs, ax)
        if grid_configs != None:
            max_peak_count = _get_max_peak_count(analysis_rois_with_sufficient_activity)
            max_len_peak_count_text = len(str(max_peak_count))
            font_size = _get_fontsize_adjusted_to_slimmest_roi(preview_image, grid_configs['window_size'], max_len_peak_count_text)
            _plot_grid(grid_configs, 'gray', 'dashed', 1, focus_area, 'cyan', 'solid', 2, ax)
            _use_grid_row_col_labels_as_axis_labels(preview_image.shape, grid_configs, ax, font_size)
        else:
            for roi in analysis_rois_with_sufficient_activity:
                plot_roi_boundaries(roi, 'magenta', 'solid', 1)
    ax.set_title(f'Total activity: {_get_total_peak_count(analysis_rois_with_sufficient_activity)}')
    return fig, ax


def _get_max_label_id_length(analysis_rois_with_sufficient_activity: List[AnalysisROI], grid_configs: Union[None, Dict[str, Any]]) -> int:
    if grid_configs != None:
        max_label_id_length = 1 + grid_configs['max_len_row_label_id'] + grid_configs['max_len_col_label_id']
    else:
        max_label_id_length = len(analysis_rois_with_sufficient_activity[0].label_id)
    return max_label_id_length


def plot_rois_with_label_id_overview(analysis_rois_with_sufficient_activity: Union[List[AnalysisROI], List[Never]],
                                     preview_image: np.ndarray,
                                     focus_area: Optional[ROI]=None,
                                     grid_configs: Optional[Dict[str, Any]]=None
                                    ) -> Tuple[Figure, Axes]:
    fig, ax = plt.subplots()
    image_row_col_dimensions = (preview_image.shape[0], preview_image.shape[1])
    blank_image = np.zeros(image_row_col_dimensions)
    ax.imshow(blank_image, cmap='gray_r')
    if len(analysis_rois_with_sufficient_activity) > 0:
        max_label_id_length = _get_max_label_id_length(analysis_rois_with_sufficient_activity, grid_configs)
        min_width_among_all_rois = _get_min_width_among_all_rois(analysis_rois_with_sufficient_activity, grid_configs)
        font_size = _get_fontsize_adjusted_to_slimmest_roi(blank_image, min_width_among_all_rois, max_label_id_length)
        for roi in analysis_rois_with_sufficient_activity:
            _plot_text_at_roi_centroid_coordinates(roi, roi.label_id, font_size, 'black', ax)
        if grid_configs != None:
            _plot_grid(grid_configs, 'black', 'solid', 1, focus_area, 'gray', 'dashed', 2, ax)
            _use_grid_row_col_labels_as_axis_labels(preview_image.shape, grid_configs, ax, font_size)
        else:
            for roi in analysis_rois_with_sufficient_activity:
                plot_roi_boundaries(roi, 'black', 'solid', 1)
    ax.set_title('IDs of ROIs with sufficient activity:')
    return fig, ax


def plot_intensity_trace_with_identified_peaks_for_individual_roi(analysis_roi: AnalysisROI) -> Figure:
    fig = plt.figure(figsize = (9, 2.67), facecolor = 'white')
    plt.plot(analysis_roi.mean_intensity_over_time, c = 'gray')
    if hasattr(analysis_roi, 'baseline'):
        plt.plot(analysis_roi.baseline, c = 'cyan')
    for peak in analysis_roi.peaks.values():
        if peak.has_neighboring_intersections == True:
            plt.plot(peak.frame_idx, peak.intensity, 'mo')
            start_idx = peak.frame_idxs_of_neighboring_intersections[0] - 1
            if start_idx < 0:
                start_idx = 0
            end_idx = peak.frame_idxs_of_neighboring_intersections[1] + 1
            plt.fill_between(np.arange(start_idx, end_idx, 1), 
                             analysis_roi.mean_intensity_over_time[start_idx : end_idx], 
                             analysis_roi.baseline[start_idx : end_idx], 
                             where = analysis_roi.mean_intensity_over_time[start_idx : end_idx] > analysis_roi.baseline[start_idx : end_idx], 
                             interpolate = True, 
                             color='yellow',
                             alpha = 0.6)
        else:
            plt.plot(peak.frame_idx, peak.intensity, 'ko')
    plt.title(f'Graph of ROI: {analysis_roi.label_id}   -   Total Activity: {analysis_roi.peaks_count}')
    plt.tight_layout()
    return fig


def export_peak_results_df_from_analysis_roi(analysis_roi: AnalysisROI) -> pd.DataFrame:
    all_peaks = [peak for peak in analysis_roi.peaks.values()]
    df_all_peak_results_single_roi = pd.DataFrame(all_peaks)
    df_all_peak_results_single_roi.drop(['has_neighboring_intersections', 'frame_idxs_of_neighboring_intersections'], axis = 'columns', inplace = True)
    df_all_peak_results_single_roi.columns = ['peak frame index', 'peak bit value', 'peak amplitude', 'peak dF/F',  'peak AUC', 'peak classification']
    df_all_peak_results_single_roi.insert(loc = 0, column = 'ROI label ID', value = f'{analysis_roi.label_id}')
    return df_all_peak_results_single_roi


def create_single_roi_amplitude_and_delta_f_over_f_results(df_all_results_single_roi: pd.DataFrame, zfill_factor: int) -> pd.DataFrame:
    rearranged_data = {'ROI label ID': [df_all_results_single_roi['ROI label ID'].iloc[0]],
                       'total peak count': [df_all_results_single_roi.shape[0]]}
    for i in range(df_all_results_single_roi.shape[0]):
        peak_idx = str(i + 1)
        peak_idx_suffix = peak_idx.zfill(zfill_factor)
        rearranged_data[f'frame index peak #{peak_idx_suffix}'] = [df_all_results_single_roi.iloc[i]['peak frame index']]
        rearranged_data[f'amplitude peak #{peak_idx_suffix}'] = [df_all_results_single_roi.iloc[i]['peak amplitude']]
        rearranged_data[f'dF/F peak #{peak_idx_suffix}'] = [df_all_results_single_roi.iloc[i]['peak dF/F']]
    return pd.DataFrame(rearranged_data)
    
    
def create_single_roi_auc_results(df_all_results_single_roi: pd.DataFrame, zfill_factor: int) -> pd.DataFrame:
    rearranged_data = {'ROI label ID': [df_all_results_single_roi['ROI label ID'].iloc[0]],
                       'total count all peaks': [df_all_results_single_roi.shape[0]],
                       'total count "singular" peaks': [df_all_results_single_roi['peak classification'].str.count('singular').sum()],
                       'total count "clustered" peaks': [df_all_results_single_roi['peak classification'].str.count('clustered').sum()],
                       'total count "isolated" peaks': [df_all_results_single_roi['peak classification'].str.count('isolated').sum()]}
    for i in range(df_all_results_single_roi.shape[0]):
        peak_idx = str(i + 1)
        peak_idx_suffix = peak_idx.zfill(zfill_factor)
        rearranged_data[f'frame index peak #{peak_idx_suffix}'] = [df_all_results_single_roi.iloc[i]['peak frame index']]
        rearranged_data[f'AUC peak #{peak_idx_suffix}'] = [df_all_results_single_roi.iloc[i]['peak AUC']]
        rearranged_data[f'classification peak #{peak_idx_suffix}'] = [df_all_results_single_roi.iloc[i]['peak classification']]
    return pd.DataFrame(rearranged_data)




"""Defines data and processing structures."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_model.ipynb.

# %% auto 0
__all__ = ['Model']

# %% ../nbs/02_model.ipynb 3
from pathlib import Path
import json
from datetime import datetime, timezone
import ipywidgets as w
import matplotlib.pyplot as plt
import inspect
import multiprocessing

# %% ../nbs/02_model.ipynb 4
from typing import Tuple, Dict, Callable, Optional, Any, List, Union
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes 

# %% ../nbs/02_model.ipynb 5
from .processing import AnalysisJob
from .input import RecordingLoaderFactory, ROILoaderFactory, RecordingLoader, ROILoader, get_filepaths_with_supported_extension_in_dirpath, FocusAreaPathRestrictions

# %% ../nbs/02_model.ipynb 6
class Model:

    def __init__(self) -> None:
        self.num_processes = multiprocessing.cpu_count()
        self.analysis_job_queue = []
        self.logs = []
        self.gui_enabled = False


    def setup_connection_to_update_infos_in_view(self, update_infos: Callable) -> None:
        self.callback_view_update_infos = update_infos
        self._check_if_gui_setup_is_completed()


    def setup_connection_to_display_results(self, show_output_screen: Callable, output: w.Output, pixel_conversion: float) -> None:
        self.callback_view_show_output_screen = show_output_screen
        self.view_output = output
        self.pixel_conversion = pixel_conversion
        self._check_if_gui_setup_is_completed()


    def _check_if_gui_setup_is_completed(self) -> None:
        checklist_expected_attributes_with_type = {'callback_view_update_infos': Callable,
                                                   'callback_view_show_output_screen': Callable,
                                                   'view_output': w.Output,
                                                   'pixel_conversion': float}
        confirmed_attributes = []
        failed_attributes = []
        for attribute_name, expected_type in checklist_expected_attributes_with_type.items():
            if hasattr(self, attribute_name) == True:
                attribute = getattr(self, attribute_name)
                if type(attribute) == expected_type:
                    confirmed_attributes.append(True)
                elif isinstance(attribute, expected_type):
                    confirmed_attributes.append(True)
                else:
                    failed_attributes.append(attribute_name)
            else:
                failed_attributes.append(attribute_name)
        if len(confirmed_attributes) == len(checklist_expected_attributes_with_type.keys()):
            self.gui_enabled = True
            self.add_info_to_logs('NA3 GUI initialization completed. Start by selecting and loading source data.', True)
        else:
            self.gui_enabled = False
            for attribute_name in failed_attributes:
                self.add_info_to_logs(f'Setup of {attribute_name} missing before GUI connection can be enabled.')


    def add_info_to_logs(self, message: str, display_in_gui: bool=False, progress_in_percent: Optional[float]=None) -> None:
        time_prefix_in_utc = datetime.now(timezone.utc).strftime('%d-%m-%y %H:%M:%S.%f')
        self.logs.append(f'{time_prefix_in_utc} (UTC): {message}')
        if (display_in_gui == True) and (self.gui_enabled == True): 
            self.callback_view_update_infos(message, progress_in_percent)


    def create_analysis_jobs(self, configs: Dict[str, Any]) -> None:
        self._ensure_data_from_previous_jobs_was_removed()
        validated_configs = self._get_configs_required_for_specific_function(configs, self._assertion_for_create_analysis_jobs)
        self.add_info_to_logs('Basic configurations for data import validated. Starting creation of analysis job(s)...', True)
        data_source_path, roi_mode, focus_area_enabled = validated_configs['data_source_path'], validated_configs['roi_mode'], validated_configs['focus_area_enabled']
        if validated_configs['batch_mode'] == True:
            all_subdir_paths_with_rec_file = self._get_all_subdir_paths_with_rec_file(data_source_path)
            all_subdir_paths_with_rec_file.sort()
            total_step_count = len(all_subdir_paths_with_rec_file)
            progress_step_size = 100 / total_step_count
            for idx, subdir_path in enumerate(all_subdir_paths_with_rec_file):
                self.add_info_to_logs(f'Starting with Job creation(s) for {subdir_path} ...', True)
                self._create_and_add_analysis_jobs_to_queue_for_single_rec(subdir_path, roi_mode, focus_area_enabled)
                self.add_info_to_logs(f'Finished Job creation(s) for {subdir_path}!', True, min((idx+1)*progress_step_size, 100.0))
        else:
            self.add_info_to_logs(f'Starting with Job creation(s) for {data_source_path} ...', True)
            self._create_and_add_analysis_jobs_to_queue_for_single_rec(data_source_path, roi_mode, focus_area_enabled)
            self.add_info_to_logs(f'Finished Job creation(s) for {data_source_path}!', True, 100.0)
        self.add_info_to_logs('All job creation(s) completed.', True, 100.0)


    def _ensure_data_from_previous_jobs_was_removed(self) -> None:
        if len(self.analysis_job_queue) > 0:
            self.add_info_to_logs('Loading of new source data. All previously created jobs & logs will be deleted.', True)
            self.analysis_job_queue = []
            self.logs = []
            self._check_if_gui_setup_is_completed()
            

    def _assertion_for_create_analysis_jobs(self, roi_mode: str, batch_mode: bool, focus_area_enabled: bool, data_source_path: Path) -> None:
        # just a convenience function to use the existing config validation and filtering methods
        pass


    def _get_all_subdir_paths_with_rec_file(self, top_level_dir_path: Path) -> List[Path]:
        rec_loader_factory = RecordingLoaderFactory()
        supported_extensions_for_recordings = rec_loader_factory.all_supported_extensions
        all_subdir_paths_that_contain_a_supported_recording_file = []
        for elem in top_level_dir_path.iterdir():
            if elem.name.startswith('.') == False:
                if elem.is_dir():
                    supported_recording_filepaths = [elem_2 for elem_2 in elem.iterdir() if elem_2.suffix in supported_extensions_for_recordings]
                    if len(supported_recording_filepaths) > 0:
                        all_subdir_paths_that_contain_a_supported_recording_file.append(elem)
        return all_subdir_paths_that_contain_a_supported_recording_file


    def _create_and_add_analysis_jobs_to_queue_for_single_rec(self, data_source_path: Path, roi_mode: str, focus_area_enabled: bool) -> None:
        recording_loader = self._get_recording_loader(data_source_path)
        if roi_mode == 'file':
            roi_loaders = self._get_all_roi_loaders(data_source_path)
        else:
            roi_loaders = None
        if focus_area_enabled == True:
            focus_area_dir_path = self._get_focus_area_dir_path(data_source_path)
            if focus_area_dir_path == None:
                analysis_job = self._create_single_analysis_job(recording_loader, roi_loaders)
                self.analysis_job_queue.append(analysis_job)
                self.add_info_to_logs(f'Successfully created a single job for {data_source_path} at queue position: #{len(self.analysis_job_queue)}.', True)
            else:
                all_focus_area_loaders = self._get_all_roi_loaders(focus_area_dir_path)
                assert len(all_focus_area_loaders) > 0, f'Focus Area analysis enabled, but no focus area ROIs could be found. Please revisit your source data and retry!'
                for idx, focus_area_loader in enumerate(all_focus_area_loaders):
                    analysis_job_with_focus_area = self._create_single_analysis_job(recording_loader, roi_loaders, focus_area_loader)
                    self.analysis_job_queue.append(analysis_job_with_focus_area)
                    job_creation_message = (f'Successfully created {idx + 1} out of {len(all_focus_area_loaders)} job(s) for {data_source_path} '
                                            f'at queue position: #{len(self.analysis_job_queue)}.')
                    self.add_info_to_logs(job_creation_message, True)
        else:
            analysis_job = self._create_single_analysis_job(recording_loader, roi_loaders)
            self.analysis_job_queue.append(analysis_job)
            self.add_info_to_logs(f'Successfully created a single job for {data_source_path} at queue position: #{len(self.analysis_job_queue)}.', True)


    def _get_recording_loader(self, data_source_path) -> RecordingLoader:
        rec_loader_factory = RecordingLoaderFactory()
        if data_source_path.is_dir() == True:
            self.add_info_to_logs(f'Looking for a valid recording file in {data_source_path}...', True)
            valid_filepaths = get_filepaths_with_supported_extension_in_dirpath(data_source_path, rec_loader_factory.all_supported_extensions, 1)
            if len(valid_filepaths) == 0:
                self.add_info_to_logs(f'Could not find any recording files of supported type at {data_source_path}!', True)
            elif len(valid_filepaths) >  1:
                recording_filepath = valid_filepaths[0]
                too_many_files_message = (f'Found more than a single recording file of supported type at {data_source_path}, i.e.: {valid_filepaths}. '
                                          f'However, only a single file was expected. NA3 continues with {recording_filepath} and will ignore the other files.')
                self.add_info_to_logs(too_many_files_message, True)
            else:
                recording_filepath = valid_filepaths[0]
                self.add_info_to_logs(f'Found recording file of supported type at: {recording_filepath}.', True)
        else:
            recording_filepath = data_source_path
            self.add_info_to_logs(f'Found recording file of supported type at: {recording_filepath}.', True)
        recording_loader = rec_loader_factory.get_loader(recording_filepath)
        return recording_loader


    def _get_all_roi_loaders(self, data_source_path: Path) -> List[ROILoader]:
        assert data_source_path.is_dir(), f'You must provide a directory as source data when using ROI mode or enabling Focus Areas. Please revisit your input data and retry.'
        roi_loader_factory = ROILoaderFactory()
        all_filepaths_with_supported_filetype_extensions = get_filepaths_with_supported_extension_in_dirpath(data_source_path, roi_loader_factory.all_supported_extensions)
        all_roi_loaders = [roi_loader_factory.get_loader(roi_filepath) for roi_filepath in all_filepaths_with_supported_filetype_extensions]
        return all_roi_loaders


    def _get_focus_area_dir_path(self, data_source_path: Path) -> Path:
        focus_area_path_restrictions = FocusAreaPathRestrictions()
        supported_dir_names = focus_area_path_restrictions.supported_dir_names
        if data_source_path.is_dir():
            source_dir_path = data_source_path
        else:
            source_dir_path = data_source_path.parent
        dirs_with_valid_name = [elem for elem in source_dir_path.iterdir() if (elem.name in supported_dir_names) & (elem.is_dir() == True)]
        if len(dirs_with_valid_name) == 0:
            no_dir_found_message = (f'You enabled Focus Area but a correspondingly named directory could not be found in {source_dir_path}. '
                                    f'Please use one of the following for the name of the directory that contains the Focus Area ROIs: {supported_dir_names}. '
                                    'In absence of such a directory, analysis is continued without using the Focus Area mode for this data.')
            self.add_info_to_logs(no_dir_found_message, True)
            focus_area_dir_path = None
        elif len(dirs_with_valid_name) > 1:
            too_many_dirs = (f'More than a single Focus Area directory was found in the following parent directory: {source_dir_path}, i.e.: '
                             f'{dirs_with_valid_name}. However, only the use of a single one that contains all your Focus Area ROIS is '
                             f'currently supported. {dirs_with_valid_name[0]} will be used for this analysis, while the other(s): {dirs_with_valid_name[1:]} '
                             'will be ignored to continue processing.')
            self.add_info_to_logs(too_many_dirs, True)
            focus_area_dir_path = dirs_with_valid_name[0]
        else:
            focus_area_dir_path = dirs_with_valid_name[0]
        return focus_area_dir_path


    def _create_single_analysis_job(self, recording_loader: RecordingLoader, roi_loaders: Union[List[ROILoader], None], focus_area_loader: Optional[ROILoader]=None) -> AnalysisJob:
        data_loaders = {'recording': recording_loader}
        if roi_loaders != None:
            data_loaders['rois'] = roi_loaders
        if focus_area_loader != None:
            data_loaders['focus_area'] = focus_area_loader
        return AnalysisJob(self.num_processes, data_loaders)
    

    def run_analysis(self, configs: Dict[str, Any]) -> None:
        sample_job_to_validate_configs = self.analysis_job_queue[0]
        validated_configs_for_analysis = self._get_configs_required_for_specific_function(configs, sample_job_to_validate_configs.run_analysis)
        validated_configs_for_result_creation = self._get_configs_required_for_specific_function(configs, sample_job_to_validate_configs.create_results)
        self.add_info_to_logs('Configurations for Analysis Settings and Result Creation validated successfully.', True)
        self.add_info_to_logs(f'Analysis Settings are:')
        for key, value in validated_configs_for_analysis.items():
            self.add_info_to_logs(f'{key}: {str(value)}')     
        self.add_info_to_logs('Starting analysis...', True)
        total_step_count = len(self.analysis_job_queue)
        progress_step_size = 100 / total_step_count
        for job_idx, analysis_job in enumerate(self.analysis_job_queue):
            self.add_info_to_logs(f'Starting to process analysis job with index #{job_idx}.')
            analysis_job.run_analysis(**validated_configs_for_analysis)
            self.add_info_to_logs(f'Analysis successfully completed. Continue with creation of results.. ')
            analysis_job.create_results(**validated_configs_for_result_creation)
            self.add_info_to_logs(f'Results successfully created at: {analysis_job.results_dir_path}')
            if self.gui_enabled == True:
                self.callback_view_show_output_screen()
                with self.view_output:
                    activity_overview_fig = analysis_job.activity_overview_plot[0]
                    activity_overview_fig.set_figheight(400 * self.pixel_conversion)
                    activity_overview_fig.tight_layout()
                    plt.show(activity_overview_fig)
            self._save_user_settings_as_json(configs, analysis_job)
            self.add_info_to_logs(f'Successfully finished processing of analysis job with index #{job_idx}. {job_idx + 1} of {total_step_count} total job(s) completed.', 
                                  True,
                                  min((job_idx+1)*progress_step_size, 100.0))
            self._save_logs_at_current_moment(analysis_job.results_dir_path)
        self.add_info_to_logs('Updating all log files to contain all logs as final step. All valid logs files will end with this message.')
        for job_idx, analysis_job in enumerate(self.analysis_job_queue):
            self._save_logs_at_current_moment(analysis_job.results_dir_path)
            

    def _get_configs_required_for_specific_function(self, all_configs: Dict[str, Any], function_to_execute: Callable) -> Dict[str, Any]:
        filtered_and_validated_configs = {}
        for expected_parameter_name in inspect.signature(function_to_execute).parameters:
            self._validate_individual_config_value_against_function_type_hints(all_configs, function_to_execute, expected_parameter_name)
            filtered_and_validated_configs[expected_parameter_name] = all_configs[expected_parameter_name]
        return filtered_and_validated_configs


    def _validate_individual_config_value_against_function_type_hints(self, all_configs: Dict[str, Any], function_to_execute: Callable, expected_parameter_name: str) -> None:
        assert expected_parameter_name in all_configs.keys(), (
            f'{function_to_execute.__name__} requires the parameter "{expected_parameter_name}", '
            f'which is not included in the configs ({list(all_configs.keys())})'
        )
        value_in_configs = all_configs[expected_parameter_name]
        expected_parameter_type = inspect.signature(function_to_execute).parameters[expected_parameter_name].annotation
        if expected_parameter_type == Path:
            assert isinstance(value_in_configs, Path), (
                f'{function_to_execute.__name__} requires the parameter "{expected_parameter_name}" to be a '
                f'pathlib.Path object. However, {value_in_configs}, which is of type '
                f'{type(value_in_configs)}, was passed.'
            )
        else:
            assert expected_parameter_type == type(value_in_configs), (
                f'{function_to_execute.__name__} requires the parameter "{expected_parameter_name}" '
                f'to be of type {expected_parameter_type}. However, {value_in_configs}, which '
                f'is of type {type(value_in_configs)}, was passed.'
            )


    def _save_logs_at_current_moment(self, results_dir_path: Path) -> None:
        filepath = results_dir_path.joinpath('logs.txt')
        with open(filepath , 'w+') as logs_file:
            for log_message in self.logs:
                logs_file.write(f'{log_message}\n')


    def _save_user_settings_as_json(self, configs: Dict[str, Any], analysis_job: AnalysisJob) -> None:
        filepath = analysis_job.results_dir_path.joinpath('user_settings.json')
        configs['recording_filepath'] = analysis_job.recording.filepath
        if analysis_job.focus_area_enabled == True:
            configs['focus_area_filepath'] = analysis_job.focus_area.filepath
        else:
            configs['focus_area_filepath'] = None
        if analysis_job.rois_source == 'file':
            for idx, roi in enumerate(analysis_job.all_rois):
                configs[f'filepath_analyzed_roi_#{idx + 1}'] = roi.filepath.as_posix()
        configs_preformatted_for_json = {}
        for key, value in configs.items():
            if isinstance(value, Path):
                configs_preformatted_for_json[key] = value.as_posix()
            else:
                configs_preformatted_for_json[key] = value
        with open(filepath, 'w+') as user_settings_json: 
            json.dump(configs_preformatted_for_json, user_settings_json)


    def preview_window_size(self, configs: Dict[str, Any]) -> Tuple[Figure, Axes]:
        job_for_preview = self.analysis_job_queue[0]
        validated_configs_for_preview = self._get_configs_required_for_specific_function(configs, job_for_preview.preview_window_size)
        preview_fig, preview_ax = job_for_preview.preview_window_size(**validated_configs_for_preview)
        return preview_fig, preview_ax
